@@ -1,18 +1,31 @@
 import weakref
 
-class BaseType(object):
+class BaseTypeByIdentity(object):
+    is_array_type = False
 
     def get_c_name(self, replace_with='', context='a C file'):
-        result = self._get_c_name(replace_with)
+        result = self.c_name_with_marker
+        assert result.count('&') == 1
+        # some logic duplication with ffi.getctype()... :-(
+        replace_with = replace_with.strip()
+        if replace_with:
+            if replace_with.startswith('*') and '&[' in result:
+                replace_with = '(%s)' % replace_with
+            elif not replace_with[0] in '[(':
+                replace_with = ' ' + replace_with
+        result = result.replace('&', replace_with)
         if '$' in result:
             from .ffiplatform import VerificationError
             raise VerificationError(
                 "cannot generate '%s' in %s: unknown type name"
-                % (self._get_c_name(''), context))
+                % (self._get_c_name(), context))
         return result
 
+    def _get_c_name(self):
+        return self.c_name_with_marker.replace('&', '')
+
     def has_c_name(self):
-        return '$' not in self._get_c_name('')
+        return '$' not in self._get_c_name()
 
     def get_cached_btype(self, ffi, finishlist, can_delay=False):
         try:
@@ -24,7 +37,13 @@ class BaseType(object):
         return BType
 
     def __repr__(self):
-        return '<%s>' % (self._get_c_name(''),)
+        return '<%s>' % (self._get_c_name(),)
+
+    def _get_items(self):
+        return [(name, getattr(self, name)) for name in self._attrs_]
+
+
+class BaseType(BaseTypeByIdentity):
 
     def __eq__(self, other):
         return (self.__class__ == other.__class__ and
@@ -33,9 +52,6 @@ class BaseType(object):
     def __ne__(self, other):
         return not self == other
 
-    def _get_items(self):
-        return [(name, getattr(self, name)) for name in self._attrs_]
-
     def __hash__(self):
         return hash((self.__class__, tuple(self._get_items())))
 
@@ -43,8 +59,8 @@ class BaseType(object):
 class VoidType(BaseType):
     _attrs_ = ()
 
-    def _get_c_name(self, replace_with):
-        return 'void' + replace_with
+    def __init__(self):
+        self.c_name_with_marker = 'void&'
 
     def build_backend_type(self, ffi, finishlist):
         return global_cache(self, ffi, 'new_void_type')
@@ -70,9 +86,9 @@ class PrimitiveType(BaseType):
         'float':              'f',
         'double':             'f',
         'long double':        'f',
-        'wchar_t':            'c',
         '_Bool':              'u',
         # the following types are not primitive in the C sense
+        'wchar_t':            'c',
         'int8_t':             'i',
         'uint8_t':            'u',
         'int16_t':            'i',
@@ -91,9 +107,7 @@ class PrimitiveType(BaseType):
     def __init__(self, name):
         assert name in self.ALL_PRIMITIVE_TYPES
         self.name = name
-
-    def _get_c_name(self, replace_with):
-        return self.name + replace_with
+        self.c_name_with_marker = name + '&'
 
     def is_char_type(self):
         return self.ALL_PRIMITIVE_TYPES[self.name] == 'c'
@@ -117,20 +131,21 @@ class BaseFunctionType(BaseType):
         self.args = args
         self.result = result
         self.ellipsis = ellipsis
-
-    def _get_c_name(self, replace_with):
-        reprargs = [arg._get_c_name('') for arg in self.args]
+        #
+        reprargs = [arg._get_c_name() for arg in self.args]
         if self.ellipsis:
             reprargs.append('...')
         reprargs = reprargs or ['void']
-        replace_with = '(%s)(%s)' % (replace_with, ', '.join(reprargs))
-        return self.result._get_c_name(replace_with)
+        replace_with = self._base_pattern % (', '.join(reprargs),)
+        self.c_name_with_marker = (
+            self.result.c_name_with_marker.replace('&', replace_with))
 
 
 class RawFunctionType(BaseFunctionType):
     # Corresponds to a C type like 'int(int)', which is the C type of
     # a function, but not a pointer-to-function.  The backend has no
     # notion of such a type; it's used temporarily by parsing.
+    _base_pattern = '(&)(%s)'
 
     def build_backend_type(self, ffi, finishlist):
         from . import api
@@ -142,9 +157,7 @@ class RawFunctionType(BaseFunctionType):
 
 
 class FunctionPtrType(BaseFunctionType):
-
-    def _get_c_name(self, replace_with):
-        return BaseFunctionType._get_c_name(self, '*'+replace_with)
+    _base_pattern = '(*&)(%s)'
 
     def build_backend_type(self, ffi, finishlist):
         result = self.result.get_cached_btype(ffi, finishlist)
@@ -157,12 +170,16 @@ class FunctionPtrType(BaseFunctionType):
 
 class PointerType(BaseType):
     _attrs_ = ('totype',)
-    
+    _base_pattern       = " *&"
+    _base_pattern_array = "(*&)"
+
     def __init__(self, totype):
         self.totype = totype
-
-    def _get_c_name(self, replace_with):
-        return self.totype._get_c_name('* ' + replace_with)
+        if totype.is_array_type:
+            extra = self._base_pattern_array
+        else:
+            extra = self._base_pattern
+        self.c_name_with_marker = totype.c_name_with_marker.replace('&', extra)
 
     def build_backend_type(self, ffi, finishlist):
         BItem = self.totype.get_cached_btype(ffi, finishlist, can_delay=True)
@@ -172,14 +189,15 @@ voidp_type = PointerType(void_type)
 
 
 class ConstPointerType(PointerType):
-
-    def _get_c_name(self, replace_with):
-        return self.totype._get_c_name(' const * ' + replace_with)
+    _base_pattern       = " const *&"
+    _base_pattern_array = "(const *&)"
 
     def build_backend_type(self, ffi, finishlist):
         BPtr = PointerType(self.totype).get_cached_btype(ffi, finishlist)
         return BPtr
 
+const_voidp_type = ConstPointerType(void_type)
+
 
 class NamedPointerType(PointerType):
     _attrs_ = ('totype', 'name')
@@ -187,41 +205,48 @@ class NamedPointerType(PointerType):
     def __init__(self, totype, name):
         PointerType.__init__(self, totype)
         self.name = name
-
-    def _get_c_name(self, replace_with):
-        return self.name + replace_with
+        self.c_name_with_marker = name + '&'
 
 
 class ArrayType(BaseType):
     _attrs_ = ('item', 'length')
+    is_array_type = True
 
     def __init__(self, item, length):
         self.item = item
         self.length = length
+        #
+        if self.length is None:
+            brackets = '&[]'
+        else:
+            brackets = '&[%d]' % self.length
+        self.c_name_with_marker = (
+            self.item.c_name_with_marker.replace('&', brackets))
 
     def resolve_length(self, newlength):
         return ArrayType(self.item, newlength)
 
-    def _get_c_name(self, replace_with):
-        if self.length is None:
-            brackets = '[]'
-        else:
-            brackets = '[%d]' % self.length
-        return self.item._get_c_name(replace_with + brackets)
-
     def build_backend_type(self, ffi, finishlist):
         self.item.get_cached_btype(ffi, finishlist)   # force the item BType
         BPtrItem = PointerType(self.item).get_cached_btype(ffi, finishlist)
         return global_cache(self, ffi, 'new_array_type', BPtrItem, self.length)
 
 
-class StructOrUnionOrEnum(BaseType):
+class StructOrUnionOrEnum(BaseTypeByIdentity):
     _attrs_ = ('name',)
     forcename = None
 
-    def _get_c_name(self, replace_with):
+    def build_c_name_with_marker(self):
         name = self.forcename or '%s %s' % (self.kind, self.name)
-        return name + replace_with
+        self.c_name_with_marker = name + '&'
+
+    def force_the_name(self, forcename):
+        self.forcename = forcename
+        self.build_c_name_with_marker()
+
+    def get_official_name(self):
+        assert self.c_name_with_marker.endswith('&')
+        return self.c_name_with_marker[:-1]
 
 
 class StructOrUnion(StructOrUnionOrEnum):
@@ -233,6 +258,7 @@ class StructOrUnion(StructOrUnionOrEnum):
         self.fldnames = fldnames
         self.fldtypes = fldtypes
         self.fldbitsize = fldbitsize
+        self.build_c_name_with_marker()
 
     def enumfields(self):
         for name, type, bitsize in zip(self.fldnames, self.fldtypes,
@@ -330,12 +356,14 @@ class StructType(StructOrUnion):
     def check_not_partial(self):
         if self.partial and self.fixedlayout is None:
             from . import ffiplatform
-            raise ffiplatform.VerificationMissing(self._get_c_name(''))
+            raise ffiplatform.VerificationMissing(self._get_c_name())
 
     def build_backend_type(self, ffi, finishlist):
         self.check_not_partial()
         finishlist.append(self)
-        return ffi._backend.new_struct_type(self.name)
+        
+        return global_cache(self, ffi, 'new_struct_type',
+                            self.get_official_name(), key=self)
 
 
 class UnionType(StructOrUnion):
@@ -343,7 +371,8 @@ class UnionType(StructOrUnion):
 
     def build_backend_type(self, ffi, finishlist):
         finishlist.append(self)
-        return ffi._backend.new_union_type(self.name)
+        return global_cache(self, ffi, 'new_union_type',
+                            self.get_official_name(), key=self)
 
 
 class EnumType(StructOrUnionOrEnum):
@@ -351,27 +380,62 @@ class EnumType(StructOrUnionOrEnum):
     partial = False
     partial_resolved = False
 
-    def __init__(self, name, enumerators, enumvalues):
+    def __init__(self, name, enumerators, enumvalues, baseinttype=None):
         self.name = name
         self.enumerators = enumerators
         self.enumvalues = enumvalues
+        self.baseinttype = baseinttype
+        self.build_c_name_with_marker()
 
     def check_not_partial(self):
         if self.partial and not self.partial_resolved:
             from . import ffiplatform
-            raise ffiplatform.VerificationMissing(self._get_c_name(''))
+            raise ffiplatform.VerificationMissing(self._get_c_name())
 
     def build_backend_type(self, ffi, finishlist):
         self.check_not_partial()
-        return ffi._backend.new_enum_type(self.name, self.enumerators,
-                                          self.enumvalues)
+        base_btype = self.build_baseinttype(ffi, finishlist)
+        return global_cache(self, ffi, 'new_enum_type',
+                            self.get_official_name(),
+                            self.enumerators, self.enumvalues,
+                            base_btype, key=self)
 
+    def build_baseinttype(self, ffi, finishlist):
+        if self.baseinttype is not None:
+            return self.baseinttype.get_cached_btype(ffi, finishlist)
+        #
+        if self.enumvalues:
+            smallest_value = min(self.enumvalues)
+            largest_value = max(self.enumvalues)
+        else:
+            smallest_value = 0
+            largest_value = 0
+        if smallest_value < 0:   # needs a signed type
+            sign = 1
+            candidate1 = PrimitiveType("int")
+            candidate2 = PrimitiveType("long")
+        else:
+            sign = 0
+            candidate1 = PrimitiveType("unsigned int")
+            candidate2 = PrimitiveType("unsigned long")
+        btype1 = candidate1.get_cached_btype(ffi, finishlist)
+        btype2 = candidate2.get_cached_btype(ffi, finishlist)
+        size1 = ffi.sizeof(btype1)
+        size2 = ffi.sizeof(btype2)
+        if (smallest_value >= ((-1) << (8*size1-1)) and
+            largest_value < (1 << (8*size1-sign))):
+            return btype1
+        if (smallest_value >= ((-1) << (8*size2-1)) and
+            largest_value < (1 << (8*size2-sign))):
+            return btype2
+        raise api.CDefError("%s values don't all fit into either 'long' "
+                            "or 'unsigned long'" % self._get_c_name())
 
 def unknown_type(name, structname=None):
     if structname is None:
         structname = '$%s' % name
     tp = StructType(structname, None, None, None)
-    tp.forcename = name
+    tp.force_the_name(name)
     return tp
 
 def unknown_ptr_type(name, structname=None):
@@ -380,11 +444,11 @@ def unknown_ptr_type(name, structname=None):
     tp = StructType(structname, None, None, None)
     return NamedPointerType(tp, name)
 
-file_type = unknown_type('FILE', '_IO_FILE')
-
-def global_cache(srctype, ffi, funcname, *args):
+def global_cache(srctype, ffi, funcname, *args, **kwds):
+    key = kwds.pop('key', (funcname, args))
+    assert not kwds
     try:
-        return ffi._backend.__typecache[args]
+        return ffi._backend.__typecache[key]
     except KeyError:
         pass
     except AttributeError:
@@ -400,8 +464,12 @@ def global_cache(srctype, ffi, funcname, *args):
         res = getattr(ffi._backend, funcname)(*args)
     except NotImplementedError as e:
         raise NotImplementedError("%r: %s" % (srctype, e))
-    ffi._backend.__typecache[args] = res
+    ffi._backend.__typecache[key] = res
     return res
 
 def pointer_cache(ffi, BType):
     return global_cache('?', ffi, 'new_pointer_type', BType)
+
+def attach_exception_info(e, name):
+    if e.args and type(e.args[0]) is str:
+        e.args = ('%s: %s' % (name, e.args[0]),) + e.args[1:]
