@@ -5,25 +5,32 @@ from . import ffiplatform
 
 class Verifier(object):
 
-    def __init__(self, ffi, preamble, tmpdir=None, ext_package=None,
-                 tag='', force_generic_engine=False, **kwds):
+    def __init__(self, ffi, preamble, tmpdir=None, modulename=None,
+                 ext_package=None, tag='', force_generic_engine=False, **kwds):
         self.ffi = ffi
         self.preamble = preamble
+        if not modulename:
+            flattened_kwds = ffiplatform.flatten(kwds)
         vengine_class = _locate_engine_class(ffi, force_generic_engine)
         self._vengine = vengine_class(self)
         self._vengine.patch_extension_kwds(kwds)
         self.kwds = kwds
         #
-        key = '\x00'.join(['1', sys.version[:3], __version__, preamble] +
-                          ffi._cdefsources)
-        if sys.version_info >= (3,):
-            key = key.encode('utf-8')
-        k1 = hex(binascii.crc32(key[0::2]) & 0xffffffff)
-        k1 = k1.lstrip('0x').rstrip('L')
-        k2 = hex(binascii.crc32(key[1::2]) & 0xffffffff)
-        k2 = k2.lstrip('0').rstrip('L')
-        modulename = '_cffi_%s_%s%s%s' % (tag, self._vengine._class_key,
-                                          k1, k2)
+        if modulename:
+            if tag:
+                raise TypeError("can't specify both 'modulename' and 'tag'")
+        else:
+            key = '\x00'.join([sys.version[:3], __version__, preamble,
+                               flattened_kwds] +
+                              ffi._cdefsources)
+            if sys.version_info >= (3,):
+                key = key.encode('utf-8')
+            k1 = hex(binascii.crc32(key[0::2]) & 0xffffffff)
+            k1 = k1.lstrip('0x').rstrip('L')
+            k2 = hex(binascii.crc32(key[1::2]) & 0xffffffff)
+            k2 = k2.lstrip('0').rstrip('L')
+            modulename = '_cffi_%s_%s%s%s' % (tag, self._vengine._class_key,
+                                              k1, k2)
         suffix = _get_so_suffix()
         self.tmpdir = tmpdir or _caller_dir_pycache()
         self.sourcefilename = os.path.join(self.tmpdir, modulename + '.c')
@@ -65,7 +72,12 @@ class Verifier(object):
         basename = os.path.basename(self.modulefilename)
         # kill both the .so extension and the other .'s, as introduced
         # by Python 3: 'basename.cpython-33m.so'
-        return basename.split('.', 1)[0]
+        basename = basename.split('.', 1)[0]
+        # and the _d added in Python 2 debug builds --- but try to be
+        # conservative and not kill a legitimate _d
+        if basename.endswith('_d') and hasattr(sys, 'gettotalrefcount'):
+            basename = basename[:-2]
+        return basename
 
     def get_extension(self):
         if not self._has_source:
@@ -90,21 +102,10 @@ class Verifier(object):
                 path = pkg.__path__
             else:
                 path = None
-            try:
-                f, filename, descr = imp.find_module(self.get_module_name(),
-                                                     path)
-            except ImportError:
+            filename = self._vengine.find_module(self.get_module_name(), path,
+                                                 _get_so_suffix())
+            if filename is None:
                 return
-            if f is not None:
-                f.close()
-            if filename.lower().endswith('.py'):
-                # on PyPy, if there are both .py and .pypy-19.so files in
-                # the same directory, the .py file is returned.  That's the
-                # case after a setuptools installation.  We never want to
-                # load the .py file here...
-                filename = filename[:-3] + _get_so_suffix()
-                if not os.path.isfile(filename):
-                    return
             self.modulefilename = filename
         self._vengine.collect_types()
         self._has_module = True
@@ -216,7 +217,11 @@ def _get_so_suffix():
     for suffix, mode, type in imp.get_suffixes():
         if type == imp.C_EXTENSION:
             return suffix
-    raise ffiplatform.VerificationError("no C_EXTENSION available")
+    # bah, no C_EXTENSION available.  Occurs on pypy without cpyext
+    if sys.platform == 'win32':
+        return ".pyd"
+    else:
+        return ".so"
 
 def _ensure_dir(filename):
     try:
