@@ -1,11 +1,20 @@
-import py, os
+import py, os, sys
 import cffi, _cffi_backend
 
+def setup_module(mod):
+    if '_cffi_backend' in sys.builtin_module_names:
+        py.test.skip("this is embedded version")
+
+BACKEND_VERSIONS = {
+    '0.4.2': '0.4',     # did not change
+    '0.7.1': '0.7',     # did not change
+    }
+
 def test_version():
     v = cffi.__version__
     version_info = '.'.join(str(i) for i in cffi.__version_info__)
     assert v == version_info
-    assert v == _cffi_backend.__version__
+    assert BACKEND_VERSIONS.get(v, v) == _cffi_backend.__version__
 
 def test_doc_version():
     parent = os.path.dirname(os.path.dirname(__file__))
@@ -13,7 +22,7 @@ def test_doc_version():
     content = open(p).read()
     #
     v = cffi.__version__
-    assert ("version = '%s'\n" % v) in content
+    assert ("version = '%s'\n" % BACKEND_VERSIONS.get(v, v)) in content
     assert ("release = '%s'\n" % v) in content
 
 def test_doc_version_file():
@@ -21,8 +30,7 @@ def test_doc_version_file():
     v = cffi.__version__
     p = os.path.join(parent, 'doc', 'source', 'index.rst')
     content = open(p).read()
-    if ("cffi/cffi-%s.tar.gz" % v) not in content:
-        py.test.skip("XXX fix the file referenced by the doc!")
+    assert ("cffi/cffi-%s.tar.gz" % v) in content
 
 def test_setup_version():
     parent = os.path.dirname(os.path.dirname(__file__))
@@ -31,3 +39,11 @@ def test_setup_version():
     #
     v = cffi.__version__
     assert ("version='%s'" % v) in content
+
+def test_c_version():
+    parent = os.path.dirname(os.path.dirname(__file__))
+    v = cffi.__version__
+    p = os.path.join(parent, 'c', 'test_c.py')
+    content = open(p).read()
+    assert (('assert __version__ == "%s"' % BACKEND_VERSIONS.get(v, v))
+            in content)
